/*
 *  client/cache.c
 *		Interface to the LysKOM services.  Also does some
 *		caching of information from the server, thus the
 *		name.
 *
 *
 *  Copyright (C) 1990	Lysator Computer Club,
 *			Linkoping University,  Sweden
 *
 *  Everyone is granted permission to copy, modify and redistribute
 *  this code, provided the people they give it to can.
 *
 *
 *  Author:	Thomas Bellman
 *		Lysator Computer Club
 *		Linkoping University
 *		Sweden
 *
 *  email:	Bellman@Lysator.LiU.SE
 *
 *
 *  Any opinions expressed in this code are the author's PERSONAL opinions,
 *  and does NOT, repeat NOT, represent any official standpoint of Lysator,
 *  even if so stated.
 */



#include <string.h>
#include <signal.h>
#include <malloc.h>

#include <compiler.h>

#include <zmalloc.h>
#include <s-string.h>

#include "config.h"
#include <misc-types.h>
#include <kom-types.h>
#include <services.h>
#include <kom-errno.h>

#include "copy.h"

#define CLIENT_CACHE

#include "error.h"

#include "misc-parser.h"
#include "cache.h"
#include "cmds.h"

#define Export



/*  ==================================================================  */
/*		    	    Global variables				*/


Export  Pers_no		  cpn	= 0;	/* Current Person Number */
Export  Conf_no		  ccn	= 0;	/* Current Conference Number */



/*  ==================================================================	*/
/*			    Local variables				*/


/*  Variables containing the cached data  */

static  Person		* this_person_stat	= NULL;
static	Conf_no		  saved_member_list_no	= 0;
static	Member_list	* saved_member_list	= NULL;
static	Pers_no		  saved_pers_stat_no	= 0;
static	Person		* saved_pers_stat	= NULL;
static	Text_no		  saved_text_no		= 0;
static	String		  saved_text		= EMPTY_STRING_i;
static	Conf_no		  saved_text_map_conf	= 0;
static	Text_list	  saved_text_map	= EMPTY_TEXT_LIST_i;
static	Membership_list	  saved_membership_list	= EMPTY_MEMBERSHIP_LIST_i;
static	Pers_no		  saved_membership_pers	= 0;
static	int		  saved_conf_names_count = -1;
static  String		**saved_conf_names	= NULL;
static	int		* saved_conf_last	= NULL;


/*  ==================================================================  */
/*		Macros and declarations of local functions.		*/


/*
 *  Lock and unlock the cache, so it won't get disturbed by
 *  signals and other strange things.  There must be a variable
 *  of type 'Signal_state' named 'save_signal' in the function
 *  calling LOCK and UNLOCK.  (And you must do the LOCK and
 *  UNLOCK in the same function.)
 */
#define LOCK	oldsigmask = sigblock (sigmask (SIGINT))
#define UNLOCK	sigsetmask (oldsigmask)



/*  Return position of conf CONF in LIST, or -1 if not present  */
Inline static  int	conf_pos (Conf_no conf, Membership_list *list);



/*  ==================================================================  */
/*	Functions to get various data from the server, possibly		*/
/*	through the cache.						*/



/*
 *  Function to handle a newly created text.
 *  Push all the info on a small queue.
 */
struct atext {
    Text_no		  text_no;
    Text_stat		  text_s;
    struct atext	* next;
};
static	struct atext 	* first = NULL;	/* The empty list is represented by */
static	struct atext	* last  = NULL;	/* first being NULL. */


/* Catches a text and puts it in the list. */
static void
catch_a_new_text(Text_no text_no, Text_stat text_stat)
{
    struct atext	* element;

    element = (struct atext *) zmalloc(sizeof(struct atext));

    element->text_no = text_no;
    element->text_s = text_stat;
    element->next = NULL;

    if (first == NULL)		/* We have an empty queue. */
    {
	first = element;
	last = element;
    } else {
	last->next = element;
	last = element;
    }
}

/* Saves a messages in a list. */
struct amessage {
    Conf_no		  to, from;
    String		  message;
    struct amessage 	* next;
};
static	struct amessage * first_message = NULL;
static	struct amessage * last_message = NULL;

static void
catch_a_directed_message(Conf_no to, Conf_no from, String message)
{
    struct amessage	* element;

    element = (struct amessage *) zmalloc(sizeof(struct amessage));

    element->to = to;
    element->from = from;
    element->message = message;
    (void) s_mark(&(element->message));
    element->next = NULL;

    if (first_message == NULL)	/* We have an empty queue. */
    {
	first_message = element;
	last_message = element;
    } else {
	last_message->next = element;
	last_message = element;
    }
}


/* Handle all asyncs. Fills in caches... Done when we can safely do so. */
Export void
handle_all_asyncs(void)
{
    struct atext	* element;

    /* Handle async messages */
    while (first_message != NULL)
    {
	struct amessage	* element;

	show_directed_message(first_message->to, 
			      first_message->from, 
			      first_message->message);

	element = first_message;
	first_message = first_message->next;
	zfree(element);
    }


    /* Handle texts */
    while (first != NULL)
    {
	Misc_info_group		  misc_data;
	const Misc_info		* misc_list;

	/* Update the last text in conf accordingly (if saved) */
	misc_list = first->text_s.misc_items;
	while ((misc_data = parse_next_misc (&misc_list,
					     first->text_s.misc_items + 
					     first->text_s.no_of_misc)).type
	       != m_end_of_list)
	{
	    Conf_no	recpt;

	    switch (misc_data.type)
	    {
	    case  m_recpt:
		recpt = misc_data.recipient;
		break;
	    case  m_cc_recpt:
		recpt = misc_data.cc_recipient;
		break;
	    default:
		recpt = 0;
	    }

	    if (recpt)		/* No recpt at this misc-position. */
	    {
		if (recpt < saved_conf_names_count /* We have saved this one */
		    && misc_data.local_no > saved_conf_last[recpt])
				/* This async is current. */
		{
		    saved_conf_last[recpt] = misc_data.local_no;
		}
	    }
	}

	element = first;
	first = first->next;
	zfree(element);
    }
}


/*
 *  Initialize the cache.  Must be called at the start of the
 *  program.
 */
Export  void
init_cache (void)

{
    register_new_text(&catch_a_new_text);
    register_directed_message(&catch_a_directed_message);
}





/*
 *  Changes what the current user is.  The new user must be logged in
 *  on the server.
 */
Export  void
cache_change_person (Pers_no	pers)

{
#define FUNCTION "cache_change_person()"

    Kom_err		  error;

    /* First, forget all in cache */
    zfree (release_person (this_person_stat));
    this_person_stat = NULL;
    zfree (release_member_list (saved_member_list));
    saved_member_list		= NULL;
    saved_member_list_no	= 0;
    zfree (release_person (saved_pers_stat));
    saved_pers_stat		= NULL;
    saved_pers_stat_no		= 0;
    s_clear (&saved_text);
    saved_text_no		= 0;
    release_text_list (&saved_text_map);
    saved_text_map		= EMPTY_TEXT_LIST;
    saved_text_map_conf		= 0;
    release_membership_list (&saved_membership_list);
    saved_membership_list	= EMPTY_MEMBERSHIP_LIST;
    saved_membership_pers	= 0;

    /* Then, get some initial chace data */
    this_person_stat = pers_stat (pers, &error);
    /* If we can't get the status for ourself, then there's not much we
     * do.  Let's just go and die then.					*/
    if (this_person_stat == NULL)
    {
	fatal (CLIENT_UNEXPECTED_SERVER_ERROR, 
	       "Can't get our own status");
    }

    /* And remember who we are */
    cpn = pers;

    /* And forget where we were */
    ccn = 0;
#undef FUNCTION
}



/*
 *  Indicate for the cache that the user has switched the current conference.
 *  Returns zero in case of error, otherwise NEW_CONF.
 */
Export  Conf_no
cache_change_conf (Conf_no	  new_conf,
		   Kom_err	* error    )

{
#define FUNCTION "cache_change_conf()"

    Success		  retval;


    retval = kom_pepsi (new_conf);
    if (retval != FAILURE)
	*error = KOM_NO_ERROR;
    else
    {
	*error = kom_errno;
	switch (kom_errno)
	{
	case  KOM_UNDEF_CONF:
	case  KOM_NOT_MEMBER:
	    return  FAILURE;
	    break;

	case  KOM_NO_CONNECT:
	case  KOM_OUT_OF_MEMORY:
	    fatal (CLIENT_SERVER_ERROR, "pepsi(%d)", new_conf);
	    break;

	default:
	    fatal (CLIENT_UNEXPECTED_SERVER_ERROR, "pepsi(%d)", new_conf);
	    break;
	}
    }   /* if (retval == FAILURE) */
    ccn = new_conf;

    return  new_conf;
#undef FUNCTION
}   /* END: cache_change_conf */



/*
 *  Get the name of a conference.  Returns EMPTY_STRING if
 *  it can't find the name.
 */
Export  String
conf_name (Conf_no	  conf,
	   Kom_err	* error)

{
    Conference		* stat;

    *error = KOM_NO_ERROR;
    if (conf == 0)
    {
	*error = KOM_CONF_ZERO;
	return  EMPTY_STRING;
    }
    if (conf <= saved_conf_names_count
	&& saved_conf_names[conf]) {
	return *s_mark (saved_conf_names[conf]) ;
    }
    stat = conf_stat(conf, error);
    release_conf(stat);

    if (saved_conf_names[conf])
	return  *s_mark (saved_conf_names[conf]);
    else			/* Failed to find the name */
	return EMPTY_STRING;
}   /* END: conf_name */



/*
 *  Get the last local_no of a conference.  Returns -1 if
 *  it can't find the conf.
 */
Export  int
conf_last (Conf_no	  conf,
	   Kom_err	* error)

{
    Conference		* stat;

    *error = KOM_NO_ERROR;
    if (conf == 0)
    {
	*error = KOM_CONF_ZERO;
	return -1;
    }
    if (conf <= saved_conf_names_count
	&& (saved_conf_last[conf] != -1)) {
	return saved_conf_last[conf];
    }
    stat = conf_stat(conf, error);
    release_conf(stat);

    return saved_conf_last[conf];
}   /* END: conf_last */



/*
 *  Return a pointer to a Conference struct for the requested
 *  conference.
 *  A NULL pointer is returned if no data about the conference
 *  can be obtained.
 */
Export  Conference *
conf_stat (Conf_no	  conf,
	   Kom_err	* error)

{
#define FUNCTION "conf_stat()"
    Conference	 	* stat;
    
    Success		  retval;

    stat = (Conference *) zmalloc( sizeof (Conference) );
    if (stat == NULL) {
	fatal (CLIENT_OUT_OF_MEMORY, "zmalloc()");
    }
    *stat = EMPTY_CONFERENCE;

    *error = KOM_NO_ERROR;
    if (conf == 0)
    {
	*error = KOM_CONF_ZERO;
	return  NULL;
    }

    retval = kom_get_conf_stat (conf, stat);
    if (retval != FAILURE)
	*error = KOM_NO_ERROR;
    else
    {
	*error = kom_errno;
	switch (*error)
	{
	case  KOM_UNDEF_CONF:
	    return  NULL;

	case  KOM_NO_CONNECT:
	    fatal (CLIENT_SERVER_ERROR, "conf_stat(%d)", conf);
	    break;

	default:
	    fatal (CLIENT_UNEXPECTED_SERVER_ERROR, "conf_stat(%d)", conf);
	    break;
	}
	return  NULL;
    }

    if (conf > saved_conf_names_count) {
	int r;

	if (!saved_conf_names) {
	    saved_conf_names = (String **) zmalloc((conf + 1) * sizeof(String *));
	    if (saved_conf_names == NULL) {
		fatal (CLIENT_OUT_OF_MEMORY, "zmalloc()");
	    }
	    saved_conf_last = (int *) zmalloc((conf + 1) * sizeof(int));
	    if (saved_conf_last == NULL) {
		fatal (CLIENT_OUT_OF_MEMORY, "zmalloc()");
	    }
	}
	saved_conf_names = (String **) zrealloc((char *) saved_conf_names,
						(conf + 1) * sizeof(String *));
	if (saved_conf_names == NULL) {
	    fatal (CLIENT_OUT_OF_MEMORY, "zrealloc()");
	}
	saved_conf_last = (int *) zrealloc((char *) saved_conf_last,
					   (conf + 1) * sizeof(int));
	if (saved_conf_last == NULL) {
	    fatal (CLIENT_OUT_OF_MEMORY, "zrealloc()");
	}
	for (r = saved_conf_names_count + 1; r <= conf; r++) {
	    saved_conf_names[r] = NULL;
	    saved_conf_last[r] = -1;
	}
	saved_conf_names_count = conf;
    }	
    if (saved_conf_names[conf])
	(void) s_release(saved_conf_names[conf]);

    (void) s_mark(&stat->name);
    saved_conf_names[conf] = &stat->name;
    saved_conf_last[conf] = stat->first_local_no + stat->no_of_texts - 1;

    return (Conference *) zuse (mark_conf(stat));
#undef FUNCTION
}   /* END: conf_stat */



/*
 *  Get the list of members for a conference.  Returns NULL if none
 *  can be found.
 */
Export  Member_list *
member_list (Conf_no	  conf,
	     Kom_err	* error)

{
#define FUNCTION "member_list()"

    Success		  retval;


    *error = KOM_NO_ERROR;
    if (conf == 0)
    {
	*error = KOM_CONF_ZERO;
	return  NULL;
    }
    else if (conf == saved_member_list_no)
	return  zuse (mark_member_list (saved_member_list));
    else
    {
	zfree (release_member_list (saved_member_list));
	saved_member_list = zmalloc (sizeof (Member_list));
	if (saved_member_list == NULL)
	{
	    fatal (CLIENT_OUT_OF_MEMORY, "zmalloc()");
	}
	*saved_member_list = EMPTY_MEMBER_LIST;
	saved_member_list_no = 0;

	retval = kom_get_members (conf, 0, MAX_CONF_NO,
				  saved_member_list);
	if (retval != FAILURE)
	    *error = KOM_NO_ERROR;
	else
	{
	    *error = kom_errno;
	    switch (*error)
	    {
	    case  KOM_UNDEF_CONF:
		return  NULL;

	    case  KOM_NO_CONNECT:
		fatal (CLIENT_SERVER_ERROR, "conf_stat(%d)");
		break;

	    default:
		fatal (CLIENT_UNEXPECTED_SERVER_ERROR, "conf_stat(%d)", conf);
		break;
	    }
	    return  NULL;
	}

	saved_member_list_no = conf;
	return  zuse (mark_member_list (saved_member_list));
    }
#undef FUNCTION
}   /* END: member_list() */
    



/*
 *  Return a pointer to a Person struct for the requested
 *  person.
 *  A NULL pointer is returned if no data about the person
 *  can be obtained.
 */
Export  Person *
pers_stat (Pers_no	  pers,
	   Kom_err	* error)

{
#define FUNCTION "pers_stat()"

    Success		  retval;


    *error = KOM_NO_ERROR;
    if (pers == 0)
    {
	*error = KOM_CONF_ZERO;	/* //////////////////// KOM_PERS_ZERO */
	return  NULL;
    }
    else if (pers == cpn)
	return  zuse (mark_person (this_person_stat));
    else if (pers == saved_pers_stat_no)
	return  zuse (mark_person (saved_pers_stat));
    else
    {
	zfree (release_person (saved_pers_stat));
	saved_pers_stat = zmalloc (sizeof (Person));
	if (saved_pers_stat == NULL)
	{
	    fatal (CLIENT_OUT_OF_MEMORY, "zrealloc()");
	}
	*saved_pers_stat = EMPTY_PERSON;
	saved_pers_stat_no = 0;

	retval = kom_get_person_stat (pers, saved_pers_stat);
	if (retval != FAILURE)
	    *error = KOM_NO_ERROR;
	else
	{
	    *error = kom_errno;
	    switch (*error)
	    {
	    case  KOM_UNDEF_PERS:
		return  NULL;

	    case  KOM_NO_CONNECT:
		fatal (CLIENT_SERVER_ERROR, "pers_stat(%d)", pers);
		break;

	    default:
		fatal (CLIENT_UNEXPECTED_SERVER_ERROR, "pers_stat(%d)", pers);
		break;
	    }
	    return  NULL;
	}

	saved_pers_stat_no = pers;
	return  zuse (mark_person (saved_pers_stat));
    }
#undef FUNCTION
}   /* END: pers_stat */



/*
 *  Copies a Text_stat and the misc-list in it from SOURCE to DEST.
 *  Both SOURCE and DEST are pointers.  The value of DEST is set.
 */
#define COPY_TEXT_STAT(source, dest)	do				\
{									\
    if (((dest) = malloc (sizeof (Text_stat))) == NULL)			\
    {									\
	fatal_client_error (CLIENT_OUT_OF_MEMORY,			\
			    this_function, "malloc() 1 failed",		\
			    (fatal_info) text);				\
	}								\
	*(dest) = *(source);						\
	(dest)->misc_items = malloc ( (dest)->no_of_misc		\
				     * sizeof (Misc_info));		\
	if ((dest)->misc_items == NULL)					\
	{								\
	    fatal_client_error (CLIENT_OUT_OF_MEMORY,			\
				this_function, "malloc() 2 failed",	\
				(fatal_info) text);			\
	}								\
	memcpy ((dest)->misc_items, (source)->misc_items,		\
		(dest)->no_of_misc * sizeof (Misc_info));		\
}  while (0)



/*
 *  Return pointer to Text_stat struct for the requested text.
 *  A NULL pointer is returned if no data about the text can be
 *  obtained.
 *  Note: no real caching is done, since the text stats tend to
 *  change rather often.
 */
Export  Text_stat *
text_stat (Text_no	  text,
	   Kom_err	* error)

{
#define FUNCTION "text_stat()"

    Success		  retval;
    Text_stat		  temp_stat	= EMPTY_TEXT_STAT;
    Text_stat		* result	= NULL;


    if (text == 0)
	return  NULL;
    else
    {
	retval = kom_get_text_stat (text, &temp_stat);
	if (retval == FAILURE)
	{
	    switch (kom_errno)
	    {
	    case  KOM_NO_SUCH_TEXT:
		return  NULL;

	    case  KOM_NO_CONNECT:
		fatal (CLIENT_SERVER_ERROR, "get_text_stat(%t)", text);
		break;

	    case  KOM_OUT_OF_MEMORY:
		fatal (CLIENT_OUT_OF_MEMORY, "get_pers_stat(%d)", text);
		break;
	    default:
		fatal (CLIENT_UNEXPECTED_SERVER_ERROR, "get_text_stat()",text);
		break;
	    }
	    return  NULL;
	}

	if ((result = zmalloc (sizeof (Text_stat))) == NULL)
	{
	    fatal (CLIENT_OUT_OF_MEMORY, "zmalloc() copying struct");
	}
	*result = temp_stat;
	result->misc_items = zmalloc ( result->no_of_misc
				      * sizeof (Misc_info));
	if (result->misc_items == NULL)
	{
	    fatal (CLIENT_OUT_OF_MEMORY, "zmalloc() copying misc");
	}
	memcpy (result->misc_items, temp_stat.misc_items,
		result->no_of_misc * sizeof (Misc_info));
	zfree (temp_stat.misc_items);

	return  result;
    }   /* text != 0 */
#undef FUNCTION
}   /* END: text_stat */



/*
 *  Return a 'String' object containing the text of the requested
 *  text, or EMPTY_STRING if the text can't be obtained.
 */
Export  String
get_text (Text_no	  text,
	  String_size	  start_char,
	  String_size	  end_char,
	  Kom_err	* error	     )

{
#define FUNCTION "get_text()"

    Success		  retval;
    String		  result	= EMPTY_STRING;


    if (text == 0)
	return  EMPTY_STRING;
    else if (text != saved_text_no)
    {
	(void)s_release (&saved_text);
	saved_text = EMPTY_STRING;
	retval = kom_get_text (text, start_char, end_char, &saved_text);
	if (retval == FAILURE)
	{
	    switch (kom_errno)
	    {
	    case  KOM_NO_SUCH_TEXT:
		return  EMPTY_STRING;

	    case  KOM_NO_CONNECT:
		fatal (CLIENT_SERVER_ERROR, "get_text(%d)", text);
		break;

	    default:
		fatal (CLIENT_UNEXPECTED_SERVER_ERROR, "get_text()", text);
		break;
	    }
	    return  EMPTY_STRING;
	}

	saved_text_no = text;
    }
    if (start_char >= s_strlen (saved_text))	/* If outside the text */
	return  EMPTY_STRING;

    if (end_char >= s_strlen (saved_text))	/* If larger than the text */
	end_char = END_OF_STRING;
    if (s_substr (&result, saved_text,
		  start_char, end_char) == FAILURE)
    {
	fatal (CLIENT_OUT_OF_MEMORY, "s_substr(%d)", text);
    }
    return  result;
#undef FUNCTION
}   /* END: get_text */



/*
 *  Check if a text with the local number LOCAL_NO currently is present
 *  in the Text_list *LIST.  LIST is a pointer to a Text_list.
 */
#define LOCAL_NO_PRESENT(local_no, list)		\
    (  (local_no) >= (list)->first_local_no		\
     && (local_no) < ( (list)->first_local_no		\
		      + (list)->no_of_texts))


/*
 *  Get the global Text_no for the text with local number LOCAL_NO
 *  in the Text_list pointed to by LIST.  The number *must* be
 *  present in the list, i e the local number must not be out of
 *  range for *LIST.
 */
#define GLOBAL_NO(local_no, list)				\
    ((list)->texts[(local_no) - (list)->first_local_no])


/*
 *  In the conference with the number CONF, find the global
 *  Text_no of a text with the local number LOCAL_NO, or 0 if it
 *  can't find it, in which case it sets 'kom_errno' to the
 *  error, or KOM_NO_ERROR if the cause for the 0 was that the
 *  text had been GC:ed, deleted or removed from the conference.
 */
Export  Text_no
map_local_to_global (Conf_no		  conf,
		     Local_text_no	  local_no,
		     Kom_err		* error	   )

{
#define FUNCTION "map_local_to_global()"

    Success			  retval;

#define LOOK_BACK					20
#define CHUNK_SIZE					128


    if (conf == 0 || local_no == 0)
    {
	kom_errno = KOM_TEXT_ZERO;
	return  0;
    }

    if (  conf == saved_text_map_conf
	     && LOCAL_NO_PRESENT (local_no, &(saved_text_map)))
    {
	return  GLOBAL_NO (local_no, &(saved_text_map));
    }
    else
    {
	retval = kom_get_map (conf, (  (local_no >= LOOK_BACK)
				     ? local_no - LOOK_BACK : 1),
			      CHUNK_SIZE, &saved_text_map);
	if (retval == FAILURE)
	{
	    switch (kom_errno)
	    {
	    case  KOM_UNDEF_CONF:
	    case  KOM_ACCESS:
	    case  KOM_NO_SUCH_LOCAL_TEXT:
		return  0;
		break;

	    case  KOM_NO_CONNECT:
		fatal (CLIENT_SERVER_ERROR, 
		       "get_map(conf=%d, local_no=%d) failed", conf, local_no);
		break;

	    default:
		fatal (CLIENT_UNEXPECTED_SERVER_ERROR,
		       "get_map(conf=%d, local_no=%d) failed", conf, local_no);
		break;
	    }
	}

	kom_errno = KOM_NO_ERROR;
	if (LOCAL_NO_PRESENT (local_no, &saved_text_map))
	    return  GLOBAL_NO (local_no, &saved_text_map);
	else
	    return  0;
    }
#undef LOOK_BACK
#undef CHUNK_SIZE
#undef FUNCTION
}   /* END: map_local_to_global */



#if 0
// *  /*
// *   *  Fetch the list of conferences the person PERS is member of,
// *   *  starting with position FIRST, and length NO_OF_CONFS.
// *   */
// *  Export  Membership_list *
// *  memberships (Pers_no	  pers,
// *  	     Kom_err	* error)
// *  
// *  {
// *      static char			* this_function	= "memberships()";
// *  
// *      Success			  retval;
// *      Membership_list		* result;
// *  
// *  
// *      if (pers == 0)
// *  	return  NULL;
// *      else if (pers != last_asked)
// *      {
// *  	retval = kom_get_membership (pers, first, no_of_confs,
// *  				     FALSE, &result);
// *  	if (retval == FAILURE)
// *  	{
// *  	    switch (kom_errno)
// *  	    {
// *  	    case  KOM_NO_CONNECT:
// *  		fatal_client_error (CLIENT_SERVER_ERROR,
// *  				    this_function, "get_person_stat()",
// *  				    (fatal_info) pers);
// *  		break;
// *  
// *  	    case  KOM_UNDEF_PERS:
// *  		return  NULL;
// *  
// *  	    default:
// *  		fatal_client_error (CLIENT_UNEXPECTED_SERVER_ERROR,
// *  				    this_function, "get_person_stat()",
// *  				    (fatal_info) 0);
// *  		break;
// *  	    }
// *  	}
// *  
// *  	saved_list.no_of_confs = temp_stat.conferences.no_of_confs;
// *  	saved_list.confs = zrealloc (saved_list.confs,
// *  				     ( saved_list.no_of_confs
// *  				      * sizeof (Membership)));
// *  	memcpy (saved_list.confs, temp_stat.conferences.confs,
// *  		saved_list.no_of_confs * sizeof (Membership));
// *  	free (temp_stat.conferences.confs);
// *      }   /* pers != last_asked */
// *  
// *      result = zmalloc (sizeof (Membership_list));
// *      result->no_of_confs = saved_list.no_of_confs;
// *      result->confs = zuse (saved_list.confs);
// *      return  result;
// *  }   /* END: memberships */
#endif




/*
 *  Find the data about person PERS:s membership in CONF and put the
 *  data in RESULT.  Returns the position in the (zero-originated)
 *  list, or -1 if not found.
 */
/* BUG? Say why we didn't find anything? */
Export  int
locate_membership (Conf_no	  conf,
		   Pers_no	  pers,
		   Membership	* result)

{
#define FUNCTION "locate_membership()"

    Success			  retval;
    int				  position;
    u_short			  next_to_fetch;

#define CHUNK_SIZE				4


    position = -1;		/* Not found yet */
    if (conf == 0 || pers == 0)
	return  -1;

    next_to_fetch = 0;
    /*  This might need some explanation...
     *  The loop fetches another chunk of the membership list from
     *  the server.  This should be done if the requested conference
     *  can't be found in the part we have in memory, and we haven't
     *  walked through the entire list.  We know we have reached the
     *  end of the list if we we didn't get as long part as we asked
     *  for.  But, if the list we remembered from last time called
     *  happens to be shorter than CHUNK_SIZE, then we shouldn't exit.
     *  We know if this is the case if 'next_to_fetch' is 0.
     */
    while (  pers != saved_membership_pers
	   || (  (position = conf_pos (conf, &saved_membership_list)) == -1
	       && (saved_membership_list.no_of_confs == CHUNK_SIZE
		   || next_to_fetch == 0)
	       ))
    {
	retval = kom_get_membership (pers, next_to_fetch, CHUNK_SIZE,
				     TRUE, &saved_membership_list);
	if (retval == FAILURE)
	{
	    switch (kom_errno)
	    {
	    case  KOM_INDEX_OUT_OF_RANGE:
	    case  KOM_UNDEF_PERS:
		return  -1;
		break;

	    case  KOM_NO_CONNECT:
		fatal (CLIENT_SERVER_ERROR, 
		       "get_membership(pers=%d, next_to_fetch=%d,chunk=%d)",
		       pers, next_to_fetch, CHUNK_SIZE);
		break;

	    default:
		fatal (CLIENT_UNEXPECTED_SERVER_ERROR,
		       "get_membership(pers=%d, next_to_fetch=%d,chunk=%d)",
		       pers, next_to_fetch, CHUNK_SIZE);
		break;
	    }
	}
	saved_membership_pers = pers;
	next_to_fetch += CHUNK_SIZE;
    }   /* while (...) */

    if (result != NULL && position >= 0)
    {
	*result = saved_membership_list.confs[position];
    }
    return  position;/* + next_to_fetch - CHUNK_SIZE; I (Linus) think. */

#undef CHUNK_SIZE
#undef FUNCTION
}   /* END: locate_membership */


/* Mark a list of texts as read in the cache and sends the message on to the
 * server.
 * Arguments as the kom_mark_as_read.
 */
Export Success
mark_as_read (Conf_no			conf_no,
	      int			len,
	      const Local_text_no     * text_arr    )
{
    int			r;
    int			position;
    Membership	      * ship;

    kom_mark_as_read (conf_no, len, text_arr);
    position = locate_membership (conf_no, cpn, NULL);

    ship = &(saved_membership_list.confs[position]);

    for ( r = 0; r < len; r++ ) {
	if (text_arr[r] <= ship->last_text_read)
	    continue;

/* BUG! This is not gonna work when we start reading in three-order. */
	if (text_arr[r] == ship->last_text_read + 1) {
	    ++(ship->last_text_read);
	}
	else {
	    int i;

	    if (ship->read_texts)
		ship->read_texts = (Local_text_no *) 
		    zrealloc ((char *) ship->read_texts, 
			     (ship->no_of_read + 1) * sizeof (Local_text_no));
	    else
		ship->read_texts = (Local_text_no *) 
		    zmalloc ((ship->no_of_read + 1) * sizeof (Local_text_no));
    
	    for (i = ship->no_of_read ; i >= 0 ; i--) {
		if (text_arr[r] > ship->read_texts[i]) {
		    ship->read_texts[i+1] = text_arr[r];
		    break;
		}
		ship->read_texts[i+1] = ship->read_texts[i];
	    }
	    if (ship->read_texts[0] > text_arr[r])
		ship->read_texts[0] = text_arr[r];

	    ++(ship->no_of_read);
	}
    }
    while (ship->no_of_read
	   && ship->last_text_read + 1 == ship->read_texts[0])
    {
	int i;

	(ship->last_text_read)++;
	for (i = 0; i < ship->no_of_read; i++)
	    ship->read_texts[i] = ship->read_texts[i+1];
	/* We could free one int's use of memory here! */
	(ship->no_of_read)--;
    }
    return OK;
}
    

    
    

/*  ==================================================================  */
/*			    Support functions				*/


/*
 *  Find the position of the conference CONF in the membership list LIST.
 *  Returns -1 if not found.
 */
Inline static  int
conf_pos (Conf_no		  conf,
	  Membership_list	* list)
{
    int		  i;

    i = 0;
    while (i < list->no_of_confs)
    {
	if (list->confs[i].conf_no == conf)
	    return  i;
	i++;
    }
    return  -1;
}   /* END: conf_pos */
